/*****************************************************************************\
*              efs - General purpose Embedded Filesystem library              *
*          --------------------- -----------------------------------          *
*                                                                             *
* Filename : sd.c                                                             *
* Revision : Initial developement                                             *
* Description : This file contains the functions needed to use efs for        *
*               accessing files on an SD-card.                                *
*                                                                             *
\*****************************************************************************/

#include "sd.h"
#include "serial.h"
#include "STM32F103.h"


euint32 OCR;


/* Kommando an SD-Karte senden und Response einholen */

euint16 sd_Command (euint8 cmd,
                    euint16 paramx,
                    euint16 paramy)
{ euint8  i;
  euint16 R;
  euint32 RR;

  Select_Card();
  if_spiDIO(0xff);
  if_spiDIO(0x40 | cmd);
  if_spiDIO((euint8) (paramx >> 8));   /* MSB of parameter x */
  if_spiDIO((euint8) (paramx));        /* LSB of parameter x */
  if_spiDIO((euint8) (paramy >> 8));   /* MSB of parameter y */
  if_spiDIO((euint8) (paramy));        /* LSB of parameter y */
  if_spiDIO(0x95);   /* Checksum (should be only valid for first command (0) */

  /* responses einholen */
  R = 0xFF;
  i = 10;
  if (cmd == 58)     /* R3 = 40 Bit */
    { while ((R & 0x80) && (i)) { R = if_spiDIO(0xFF); --i; }
      RR = if_spiDIO(0xFF);
      RR = (RR<<8) | if_spiDIO(0xFF);
      RR = (RR<<8) | if_spiDIO(0xFF);
      RR = (RR<<8) | if_spiDIO(0xFF);
      OCR = RR;
    }
  else
  { if (cmd == 13)
    { while ((R & 0x80) && (i)) { R = if_spiDIO(0xFF); --i; }
      R = (R<<8) | if_spiDIO(0xFF);
    }
    else
    { while ((R & 0x80) && (i)) { R = if_spiDIO(0xFF); --i; } }
  }
  if (!i) R = 0xFFFF;
  if ((cmd==9)||(cmd==10)||(cmd==17)||(cmd==18)||(cmd==24)||(cmd==25)) return R;
  Unselect_Card();
  return R;
}



void sd_Resp8bError (euint8 value)
{
#ifdef DEBUG
 if (value & 0x40) V24_StrOut("bad Argument.\n");
 if (value & 0x20) V24_StrOut("bad Address.\n");
 if (value & 0x10) V24_StrOut("erase seq error.\n");
 if (value & 0x08) V24_StrOut("bad CRC.\n");
 if (value & 0x04) V24_StrOut("illegal command.\n");
 if (value & 0x02) V24_StrOut("erase reset (see SanDisk docs p5-13).\n");
 if (value & 0x01) V24_StrOut("card is idle.\n");
#endif
}




/* WAIT ?? -- FIXME
 * CMDWRITE
 * WAIT
 * CARD RESP
 * WAIT
 * DATA BLOCK OUT
 *      START BLOCK
 *      DATA
 *      CHKS (2B)
 * BUSY...
 */

esint8 sd_writeSector (euint32 address,
		       euint8* buf)
{ euint32 place;
  euint16 i;
  euint16 t=0;

  place = 512*address;
  sd_Command (CMDWRITE, (euint16) (place >> 16), (euint16) place);
  if_spiDIO(0xfe);         /* Start block */
  for(i=0;i<512;i++)
	if_spiDIO(buf[i]); /* Send data */
  if_spiDIO(0xff);         /* Checksum part 1 */
  if_spiDIO(0xff);         /* Checksum part 2 */
  if_spiDIO(0xff);
  while(if_spiDIO(0xff)!=0xff) t++;
  return(0);
}



/* WAIT ?? -- FIXME
 * CMDCMD
 * WAIT
 * CARD RESP
 * WAIT
 * DATA BLOCK IN
 * 		START BLOCK
 * 		DATA
 * 		CHKS (2B)
 */


esint8 sd_readSector (euint32 address,
		      euint8* buf,
		      euint16 len)
{ euint8 cardresp;
  euint8 firstblock;
  euint8 c;
  euint16 fb_timeout;
  euint32 i;
  euint32 place;

  place = 512*address;
  fb_timeout = 0xffff;
  cardresp = sd_Command(CMDREAD, (euint16) (place >> 16), (euint16) place);

	/* Wait for startblock */
  Select_Card();
  do firstblock = if_spiDIO(0xff);
    while (firstblock!=0xfe && fb_timeout--);

  if(cardresp!=0x00 || firstblock!=0xfe)
  { sd_Resp8bError(firstblock);
    return(-1);
  }

  for(i=0;i<512;i++)
  { c = if_spiDIO(0xff);
    if (i<len) buf[i] = c;
  }

	/* Checksum (2 byte) - ignore for now */
  if_spiDIO(0xff);
  if_spiDIO(0xff);
  Unselect_Card();
  return(0);
}



/* calculates size of card from CSD
 (extension by Martin Thomas, inspired by code from Holger Klabunde)
 */

esint8 sd_getDriveSize (euint32* drive_size )
{ euint8 cardresp, i, by;
  euint8 firstblock;
  euint16 fb_timeout;
  euint8 iob[16];
  euint16 c_size, c_size_mult, read_bl_len;

  cardresp = sd_Command(CMDREADCSD, 0, 0);
  fb_timeout = 0xffff;
  do { firstblock = if_spiDIO(0xff);
     } while (firstblock!=0xfe && fb_timeout--);

  if(cardresp!=0x00 || firstblock!=0xfe)
  { sd_Resp8bError(firstblock);
    return(-1);
  }

  for( i=0; i<16; i++) { iob[i] = if_spiDIO(0xff);  }

  if_spiDIO(0xff);
  if_spiDIO(0xff);

  c_size = iob[6] & 0x03; /* bits 1..0 */
  c_size <<= 10;
  c_size += (euint16)iob[7]<<2;
  c_size += iob[8]>>6;

  by = iob[5] & 0x0F;
  read_bl_len = 1;
  read_bl_len <<= by;

  by = iob[9] & 0x03;
  by <<= 1;
  by += iob[10] >> 7;

  c_size_mult = 1;
  c_size_mult <<= (2+by);

  *drive_size = (euint32)(c_size+1) * (euint32)c_size_mult * (euint32)read_bl_len;



 return 0;
}

